'''Tests about gzippy.'''

import unittest

from six.moves import cStringIO as StringIO

from gzippy.header import Header


class HeaderTest(unittest.TestCase):
    '''Tests about the gzippy header.'''

    def setUp(self):
        self.header = Header(
            flags=(Header.FEXTRA | Header.FNAME | Header.FCOMMENT),
            mtime=1234, 
            extra_flags=3,
            os_flags=1,
            extra='extra',
            name='name',
            comment='comment')

    def test_round_trip(self):
        '''Can round-trip headers.'''
        fout = StringIO()
        self.header.write(fout)
        self.assertEqual(Header.read(StringIO(fout.getvalue())), self.header)

    def test_invalid_magic(self):
        '''Raises an exception if the magic is wrong.'''
        with self.assertRaises(IOError):
            Header.read(StringIO('Not the right magic'))

    def test_invalid_method(self):
        '''Only the deflate method is supported.'''
        with self.assertRaises(IOError):
            Header.read(StringIO('\x1F\x8B\xFF'))

    def test_multipart_not_supported(self):
        '''Raises an exception if the header has the multibyte flag.'''
        # This is a very basic header, with just FMULTI set
        serialized = StringIO('\x1F\x8B\x08\x02\xD2\x04\x00\x00\x03\x01')
        with self.assertRaises(ValueError):
            Header.read(serialized)

    def test_repr(self):
        '''Exercises the repr function.'''
        parts = [
            '<Gzip.Header',
            'flags=28,',
            'mtime=1234,',
            'extra_flags=3,',
            'os_flags=1,',
            'extra=extra,',
            'name=name,',
            'comment=comment>'
        ]
        self.assertEqual(repr(self.header), ' '.join(parts))



'''Classes to help with reading and wrigin headers.'''

import struct
import time
import zlib


# pylint: disable=too-many-instance-attributes
class Header(object):
    '''Read, write, hold Gzip file headers.'''

    # Constatnts
    MAGIC = '\x1F\x8B'
    METHOD = chr(zlib.DEFLATED)

    # Flags
    FTEXT = 0x1
    FMULTI = 0x2
    FEXTRA = 0x4
    FNAME = 0x8
    FCOMMENT = 0x10

    # The attributes for this class
    __slots__ = (
        'flags',
        'mtime',
        'extra_flags',
        'os_flags',
        'extra',
        'name',
        'comment',
    )

    # pylint: disable=too-many-arguments
    def __init__(self, flags=0, mtime=None, extra_flags=2, os_flags=255, extra=None,
                 name=None, comment=None):
        self.flags = flags
        self.mtime = mtime or int(time.time())
        self.extra_flags = extra_flags
        self.os_flags = os_flags
        # TODO(dan): make extra access like a dictionary
        self.extra = extra
        self.name = name
        self.comment = comment

    @classmethod
    def read_null_terminated(cls, fobj):
        '''Read a null-terminated string.'''
        result = ''
        while True:
            char = fobj.read(1)
            if not char or char == '\x00':
                break
            result += char
        return result

    @classmethod
    def read(cls, fobj):
        '''Read in headers from a file object.'''
        if fobj.read(2) != cls.MAGIC:
            raise IOError('Invalid magic.')
        if fobj.read(1) != cls.METHOD:
            raise IOError('Unknown compression method')
        flags = ord(fobj.read(1))
        mtime = struct.unpack('<L', fobj.read(4))[0]
        extra_flags = ord(fobj.read(1))
        os_flags = ord(fobj.read(1))

        extra = None
        if flags & cls.FEXTRA:
            length = struct.unpack('<H', fobj.read(2))[0]
            extra = fobj.read(length)

        name = None
        if flags & cls.FNAME:
            raw = cls.read_null_terminated(fobj)
            name = raw.decode('latin-1')

        comment = None
        if flags & cls.FCOMMENT:
            comment = cls.read_null_terminated(fobj).decode('latin-1')

        if flags & cls.FMULTI:
            raise ValueError('Multipart gzips not supported.')

        return cls(flags, mtime, extra_flags, os_flags, extra, name,
                   comment)

    def write(self, fobj):
        '''Write this header to a file object.'''
        self.flags = (
            (self.FNAME if (self.name is not None) else 0) |
            (self.FEXTRA if (self.extra is not None) else 0) |
            (self.FCOMMENT if (self.comment is not None) else 0))

        fobj.write(self.MAGIC)                      # magic header
        fobj.write(self.METHOD)                     # compression method
        fobj.write(chr(self.flags))                 # flags
        fobj.write(struct.pack('<L', self.mtime))   # Modification time
        fobj.write(chr(self.extra_flags))           # eXtra FLags (max compression)
        fobj.write(chr(self.os_flags))              # Operating system (unknown)

        if (self.flags & self.FEXTRA) and self.extra:
            fobj.write(struct.pack('<H', len(self.extra)))
            fobj.write(self.extra)

        if (self.flags & self.FNAME) and self.name:
            fobj.write(self.name)
            fobj.write('\000')

        if (self.flags & self.FCOMMENT) and self.comment:
            fobj.write(self.comment)
            fobj.write('\000')

    def __repr__(self):
        joined = ', '.join(
            '%s=%s' % (attr, getattr(self, attr)) for attr in self.__slots__)
        return '<Gzip.Header %s>' % joined

    def __eq__(self, other):
        return all(
            getattr(self, attr) == getattr(other, attr) for attr in self.__slots__)



'''Tests about gzippy.files.'''

import os
import subprocess
import unittest

import mock

import gzippy
import gzippy.header

from . import scratch_file, scratch_dir

class FilesTest(unittest.TestCase):
    '''Tests about gzippy'''

    def test_round_trip(self):
        '''Can round-trip content.'''
        content = 'This is some test content.'
        with scratch_file('example.gz') as path:
            with gzippy.open(path, 'wb') as fout:
                fout.write(content)

            with gzippy.open(path, 'rb') as fin:
                self.assertEqual(fin.read(), content)

    def test_unsupported_incremental_reads(self):
        '''Incremental reads are as-of-yet unsupported.'''
        content = 'This is some test content.'
        with scratch_file('example.gz') as path:
            with gzippy.open(path, 'wb') as fout:
                fout.write(content)

            with self.assertRaises(IOError):
                with gzippy.open(path, 'rb') as fin:
                    fin.read(10)

    def test_gzip_compatible(self):
        '''Output compatible with the gzip command-line utility.'''
        content = 'This is some test content.'
        kwargs = {
            'flags': 0,
            'mtime': 1234,
            'name': 'name',
            'comment': 'comment'
        }
        with scratch_dir() as path:
            zipped = os.path.join(path, 'example.gz')
            unzipped = os.path.join(path, 'example')
            with gzippy.GzipWriter.open(zipped, **kwargs) as fout:
                fout.write(content)

            subprocess.check_call(['gunzip', zipped], stderr=subprocess.STDOUT)

            with open(unzipped) as fin:
                self.assertEqual(fin.read(), content)

    def test_lines(self):
        '''Can read the file line by line.'''
        parts = ['some\n', 'lines\n', 'in\n', 'a\n', 'file']
        content = ''.join(parts)
        with scratch_file('example.gz') as path:
            with gzippy.open(path, 'wb') as fout:
                fout.write(content)

            with gzippy.open(path) as fin:
                self.assertEqual(list(fin), parts)

    def test_lines_consolidation(self):
        '''Consolidates lines across multiple chunks.'''
        parts = ['some\n', 'lines\n', 'in\n', 'a\n', 'file']
        chunks = ['so', 'm', 'e\nlines\n', 'i', 'n', '\n', 'a\nfile']
        content = ''.join(chunks)
        with scratch_file('example.gz') as path:
            with gzippy.open(path, 'wb') as fout:
                fout.write(content)

            with gzippy.open(path) as fin:
                with mock.patch.object(fin, 'chunks', iter(chunks)):
                    self.assertEqual(list(fin), parts)

    def test_reader_crc_mismatch(self):
        '''Raises an exception when the crc doesn't match.'''
        with scratch_file('example.gz') as path:
            with gzippy.open(path, 'wb') as fout:
                fout.write('This is some test content.')

            # Rewrite the last eight bytes
            with open(path, 'r+b') as fout:
                fout.seek(-8, 2)
                fout.write('\xFF' * 8)

            with self.assertRaises(IOError):
                with gzippy.open(path) as fin:
                    fin.read()

    def test_reader_size_mismatch(self):
        '''Raises an exception when the size doesn't match.'''
        with scratch_file('example.gz') as path:
            with gzippy.open(path, 'wb') as fout:
                fout.write('This is some test content.')

            # Rewrite the last four bytes
            with open(path, 'r+b') as fout:
                fout.seek(-4, 2)
                fout.write('\xFF' * 4)

            with self.assertRaises(IOError):
                with gzippy.open(path) as fin:
                    fin.read()



'''Gzip file interactions.'''

import contextlib

from six.moves import builtins

from .reader import GzipReader
from .writer import GzipWriter


def from_file(fobj):
    '''Open a Gzip file from a file object.'''
    if '+' in fobj.mode:
        raise ValueError('Simultaneous reads and writes not supported.')
    elif 'r' in fobj.mode:
        return GzipReader(fobj)
    elif 'w' in fobj.mode:
        return GzipWriter(fobj)
    else:
        raise ValueError('Mode not supported')


# pylint: disable=redefined-builtin
@contextlib.contextmanager
def open(path, mode='rb'):
    '''Open a Gzip file.'''
    with builtins.open(path, mode) as raw:
        fobj = from_file(raw)
        try:
            yield fobj
        finally:
            fobj.close()



'''Tests about the gzippy top-level functions.'''

import unittest

import gzippy

from . import scratch_file


class GzippyTest(unittest.TestCase):
    '''Tests about the gzippy top-level functions.'''

    def test_open_with_plus(self):
        '''Opening with r+ is not allowed.'''
        with scratch_file('example.gz') as path:
            with open(path, 'w+') as fout:
                pass

            with self.assertRaises(ValueError):
                with gzippy.open(path, 'r+') as fin:
                    pass

    def test_open_with_append(self):
        '''Opening in append mode is not allowed.'''
        with scratch_file('example.gz') as path:
            with open(path, 'w+') as fout:
                pass            

            with self.assertRaises(ValueError):
                with gzippy.open(path, 'ab') as fout:
                    pass



'''A reader of gzip files.'''

import struct
import zlib

from .header import Header


class GzipReader(object):
    '''A reader of gzip files.'''

    def __init__(self, fobj):
        self.raw = fobj
        self.crc = zlib.crc32('') & 0xFFFFFFFF
        self.size = 0
        self.decompressobj = zlib.decompressobj(-zlib.MAX_WBITS)
        self.header = Header.read(self.raw)
        self.chunks = iter(self.chunks_generator())

    def chunks_generator(self):
        '''Generator of decompressed data.'''
        buf = self.raw.read(4096)
        while buf:
            uncompressed = self.decompressobj.decompress(buf)
            self.size += len(uncompressed)
            self.crc = zlib.crc32(uncompressed, self.crc) & 0xFFFFFFFF
            yield uncompressed
            buf = self.raw.read(4096)

        # Yield the last flushed block
        uncompressed = self.decompressobj.flush()
        self.size += len(uncompressed)
        self.crc = zlib.crc32(uncompressed, self.crc) & 0xFFFFFFFF
        yield uncompressed

        # Check the CRC and size
        crc, size = struct.unpack('<II', self.decompressobj.unused_data[-8:])
        if crc != self.crc:
            raise IOError('CRC mismatch: %s (read) != %s (computed)' % (crc, self.crc))
        if size != self.size:
            raise IOError('Size mismatch: %s (read) != %s (computed)' % (size, self.size))

    def lines(self):
        '''Generator of the lines in a file.'''
        last = ''
        for chunk in self.chunks:
            chunk = last + chunk
            start = 0
            index = chunk.find('\n')
            while index != -1:
                yield chunk[start:index+1]
                start, index = index + 1, chunk.find('\n', index + 1)

            last = chunk[start:]

        yield last

    def __iter__(self):
        '''Generator of the lines in this file.'''
        return self.lines()

    def read(self, size=-1):
        '''Read data from the file.'''
        if size == -1:
            return ''.join(self.chunks)
        else:
            raise IOError('Incremental reads not supported.')

    def close(self):
        '''Close the file.'''
        self.raw.close()



#! /usr/bin/env python

from distutils.core import setup

setup(name           = 'gzippy',
    version          = '0.1.2',
    description      = 'Gzip file utility.',
    url              = 'http://github.com/seomoz/gzippy',
    author           = 'Dan Lecocq',
    author_email     = 'dan@moz.com',
    packages = [
        'gzippy'
    ],
    package_dir = {
        'gzippy': 'gzippy'
    },
    install_requires = [
    ],
    tests_require    = [
        'coverage',
        'nose',
        'mock',
        'pylint',
        'rednose'
    ]
)



'''A writer of gzip files.'''

import contextlib
import struct
import zlib

from .header import Header


class GzipWriter(object):
    '''A writer of gzip files.'''

    @classmethod
    @contextlib.contextmanager
    def open(cls, path, *args, **kwargs):
        '''Open the provided path.'''
        with contextlib.closing(cls(open(path, 'wb'), *args, **kwargs)) as fout:
            yield fout

    def __init__(self, fobj, compresslevel=9, **kwargs):
        self.raw = fobj
        self.crc = zlib.crc32('') & 0xFFFFFFFF
        self.size = 0
        self.compressobj = zlib.compressobj(compresslevel,
                                            zlib.DEFLATED,
                                            -zlib.MAX_WBITS,
                                            zlib.DEF_MEM_LEVEL,
                                            0)

        # Write the header
        self.header = Header(**kwargs)
        self.header.write(self.raw)

    def close(self):
        '''Close the file.'''
        try:
            self.raw.write(self.compressobj.flush(zlib.Z_FINISH))
            self.raw.write(struct.pack('<II', self.crc, self.size))
            self.raw.flush()
        finally:
            self.raw.close()

    def write(self, data):
        '''Write data to the file.'''
        self.raw.write(self.compressobj.compress(data))
        self.size += len(data)
        self.crc = zlib.crc32(data, self.crc) & 0xFFFFFFFF
